import os
import numpy as np
import rasterio
from rasterio.merge import merge
from rasterio.mask import mask
from rasterio.warp import calculate_default_transform, reproject, Resampling
import datetime
import shutil
import fiona

def UTM_to_WGS84(file_to_proj, temp_dir):
    with rasterio.open(file_to_proj) as src:
            # Define the target CRS (WGS84)
            dst_crs = 'EPSG:4326'

            # Calculate the transform and new dimensions
            transform, width, height = calculate_default_transform(
                src.crs, dst_crs, src.width, src.height, *src.bounds)

            # Initialize the destination dataset
            dst_meta = src.meta.copy()
            dst_meta.update({
                'crs': dst_crs,
                'transform': transform,
                'width': width,
                'height': height
            })
            now = datetime.datetime.now()
            time_str = now.strftime("%M%S")
            reprojected_file = os.path.join(temp_dir, f'projected_{time_str}_{os.path.basename(file_to_proj)}')
            #print(reprojected_file)
            with rasterio.open(reprojected_file, 'w', **dst_meta) as dst:
                for i in range(1, src.count + 1):
                    reproject(
                        source=rasterio.band(src, i),
                        destination=rasterio.band(dst, i),
                        src_transform=src.transform,
                        src_crs=src.crs,
                        dst_transform=transform,
                        dst_crs=dst_crs,
                        resampling=Resampling.nearest,
                        dst_nodata=-9999
                    )
    return reprojected_file

def convert_nodata_to_value(src_file, nodata_value, new_value):
    with rasterio.open(src_file) as src:
        # Read the source raster as array
        array = src.read(1)
        # Convert NoData values to the specified integer value
        array[array == nodata_value] = new_value

        # Copy metadata and update NoData value
        out_meta = src.meta.copy()
        out_meta.update({
            'nodata': new_value,
            'dtype': 'int32'  # Ensure the dtype can handle the new integer values
        })
        # Save the new raster
        now = datetime.datetime.now()
        time_str = now.strftime("%M%S")
        dst_file = os.path.join(temp_dir, f'temp_anan_{time_str}.tif')
        with rasterio.open(dst_file, 'w', **out_meta) as dst:
            dst.write(array, 1)

    return dst_file

DirOut_eta = os.path.join(r'W:\Users\Stefanie\IWMI\Awash_basin\L8_9_52%_169-166_051-054\data\merge_clip')
if not os.path.exists(DirOut_eta):
    os.makedirs(DirOut_eta)
#wgsfile = r'W:\Projects\VegET_FEWS\output_data\Operational_Runs\Africa\lwrsi_moving_9dek\lwrsi_9dek_2024101.tif'

for year in range(2020, 2025):
    #year = '2022'  #input('Enter year:')
    year = str(year)
    for mon in range(1, 13):
        #temp folder cleanup/setup
        temp_dir = os.path.join(DirOut_eta, f'temp{year}{mon}')
        # if os.path.exists(temp_dir):
        #     shutil.rmtree(temp_dir)
        #     os.makedirs(temp_dir)
        # else:
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)

        src_files = []
        root_dir = r'W:\Users\Stefanie\IWMI\Awash_basin\L8_9_52%_169-166_051-054\data'
        tar_name = f'eta_{year}{mon:02d}.tif'
        # List of source files
        for dirpath, dirnames, filenames in os.walk(root_dir):
            for filename in filenames:
                if filename == tar_name:
                    file_name = os.path.join(dirpath, filename)
                    #reproject to WGS84
                    pro_file = UTM_to_WGS84(file_to_proj=file_name, temp_dir=temp_dir)
                    arrfile = convert_nodata_to_value(src_file=pro_file, nodata_value=-9999, new_value=0)
                    src_files.append(arrfile)
        print(f'length of list = {len(src_files)} -->', f' 1. file is: {src_files[0]}')

        src_files_to_mosaic = [rasterio.open(fp) for fp in src_files]
        mosaic, out_trans = merge(src_files_to_mosaic, method='max')

        out_meta = src_files_to_mosaic[0].meta.copy()
        out_meta.update({
            "driver": "GTiff",
            "height": mosaic.shape[1],
            "width": mosaic.shape[2],
            "count": mosaic.shape[0],
            "transform": out_trans
        })

        etaprj_out = os.path.join(DirOut_eta, f'eta_{year}{mon:02d}mosaic.tif')
        with rasterio.open(etaprj_out, 'w', **out_meta) as des:
            #des.write(mosaic)
            arr = np.where(mosaic < 0, 0, mosaic)
            arr = np.where(arr > 555, 555, arr)
            des.write(arr[0], 1)
        print(f'Saved merged Eta raster.')

        shapefile_path = r'W:\Users\Stefanie\IWMI\Awash_basin\awash_basin_wgs84.shp'
        with fiona.open(shapefile_path, 'r') as shapefile:
            shapes = [feature["geometry"] for feature in shapefile]
            #print(shapes)

        with rasterio.open(etaprj_out) as ds:
            out_image, out_transform = mask(ds, shapes, crop=True)
            out_image = np.where(out_image == ds.nodata, np.nan, out_image)
            out_meta = ds.meta.copy()
            out_meta.update({
                "driver": "GTiff",
                "height": out_image.shape[1],
                "width": out_image.shape[2],
                "transform": out_transform,
                "nodata": np.nan,
                "dtype": 'float32'  # Ensure the dtype can handle NaN
            })
        eta_out = os.path.join(DirOut_eta, f'eta_{year}{mon:02d}.tif')
        with rasterio.open(eta_out, 'w', **out_meta) as dest:
            dest.write(out_image)
        print(f'Saved final Eta raster.')