# ------------------------------------
# script: 1_postprocess_espa_files.py
# Info: Unzips and renames the downloaded ESPA tar.gz files
# -------------------------------------

# Import Modules
import datetime as dt
import glob
import os
import tarfile
import numpy as np
import rasterio
from datetime import datetime



# https://www.bing.com/search?pglt=131&q=How+to+use+Landsat+QA+band+in+python+for+cloud+masking&PC=U531&cvid=59cad6ba2324410d9279b586ae0849a6&FORM=ANNAB1&showconv=1
# https://gis.stackexchange.com/questions/449118/how-can-i-use-landsat-qa-band-to-mask-clouds-collection-2-level-1
def cloudmask(qa, date, pr, landsat):

    # open file for geoproperties to use for outputting raster
    with rasterio.open(qa) as src:
        #print(src.shape)
        outmeta = src.profile
        # outmeta.update(dtype=rasterio.uint8,
        #               count=1,
        # outmeta.update(crs=f'epsg:4326',
        #               compress='lzw')

        qa = src.read()  # read in as array
        #print(qa.shape)
        qa = qa.astype(int)  # convert to int array
        #print(type(qa), qa.shape)
        dilated = np.bitwise_and(qa, 2)  # change value to bit
        #print(dilated.shape)
        cloud = np.bitwise_and(qa, 8)
        cloudshadow = np.bitwise_and(qa, 16)
        snow = np.bitwise_and(qa, 32)
        if landsat == '8' or landsat == '9':
            cirrus = np.bitwise_and(qa, 4)
            group = [dilated, cirrus, cloud, cloudshadow, snow]
        else:
            group = [dilated, cloud, cloudshadow, snow]

        max_mask = np.maximum.reduce(group)
        #print(max_mask.shape)
        mask = np.where(max_mask == 0, 0, 1)  # clear pixels = 0, cloud pixels = 1
        #print(mask)
        #print(mask[np.isnan(mask)])  # = 1
        #print(mask.shape)

    # save cloud mask
    outfile = os.path.join(os.getcwd(), f'cmask_{pr}_{date}.tif')
    with rasterio.open(outfile, 'w', **outmeta) as wrast:
        wrast.write(mask)

    return mask


# unzip tar.gz files
def unziptar(gzfiles, tempdir):
    for gz in gzfiles:
        name = os.path.basename(gz)[4:18]   #LC08 02602920230108 02T1-SC20240507180436.tar.gz
        #landsat = os.path.basename(gz)[3:4]
        scratch = os.path.join(tempdir, 'espa_' + name)
        if not os.path.exists(scratch):
            os.makedirs(scratch)
        with tarfile.open(gz, 'r:gz') as tar:
            os.chdir(scratch)
            for member in tar.getmembers():
                tar.extract(member, scratch)

def unziptar_unix(gzfiles, tempdir):
    for gz in gzfiles:
        name = os.path.basename(gz)[4:18]   #LC08 02602920230108 02T1-SC20240507180436.tar.gz
        #landsat = os.path.basename(gz)[3:4]
        scratch = os.path.join(tempdir, 'espa_' + name)
        if not os.path.exists(scratch):
            os.makedirs(scratch)
        with tarfile.open(gz, 'r:gz') as tar:
            os.chdir(scratch)
            tar.extractall(path=scratch)


if __name__ == '__main__':

# Unzipping tar.gz files
    # find all the tar.gz files in the folder 'filedir'
    filedir = input('Enter folder path where tar files are located: ')
    gzfiles = sorted(glob.glob(filedir +os.sep+ '*.gz'))
    print(f'There are {len(gzfiles)} tar.gz files in the folder.')

    # change directory to 'filedir' and unzip the tar.gz files to folder 'unzipped'
    os.chdir(filedir)
    #print(os.getcwd())
    dir_one_up = os.path.normpath(os.getcwd() + os.sep + os.pardir)
    # unzip function
    unzipdir = os.path.join(dir_one_up, 'unzipped')
    print(f'Unzipping tar.gz files to directory: {unzipdir}')
    unziptar(gzfiles, tempdir=unzipdir)

# Scaling and applying cloud masking to ET raster
    # Uses the ETf and ETa files from the unzipped folder are scaled and cloud-mask improved
    # and saved in the designated folders organized by path/row.
    # The cloud-masking adjusts for dilated clouds,cirrus clouds,cloud shadows, and snow.
    etfiledir = os.path.join(dir_one_up, 'data')
    print(f'Creating cloud-masked ETf and ETa data in directory: {etfiledir}')
    dir_contents2 = sorted(os.listdir(unzipdir))
    for i in dir_contents2:
        actdir = unzipdir + os.sep + i
        path_row = f'p{i[6:8]}r{i[9:11]}'
        for folder in ['ETf', 'ETa']:
            if not os.path.exists(os.path.join(etfiledir, path_row, folder)):
                os.makedirs(os.path.join(etfiledir, path_row, folder))
        startTime = dt.datetime.now()
        rasters = sorted(glob.glob(actdir + os.sep + '*.tif'))
        print(rasters)
        first = rasters[0]
        name2 = os.path.basename(first)[:41]
        landsat = name2[3:4]

        date = str(os.path.basename(name2)[17:25])
        print('working on:', date)

        # scale factor for ET rasters
        etf_scale_factor = 0.0001
        eta_scale_factor = 0.001

        # defining the different ET rasters and the Quality band (qa)
        etfband = os.path.join(actdir, name2+'ETf.tif')
        etaband = os.path.join(actdir, name2+'ETa.tif')
        qaband = os.path.join(actdir, name2+'QA_PIXEL.tif')

        #Create cloud mask with function
        cmask = cloudmask(qaband, date, path_row, landsat)
        #print(cmask)

        # ETf raster
        # open file for geoproperties to use for outputting raster
        with rasterio.open(etfband) as src:
            outmeta = src.meta
            outmeta.update(dtype=rasterio.float32, compress='lzw')
            #              crs=f'epsg:4326',
            #              compress='lzw')

            #apply scalar and cloud mask
            scaled_etf = src.read()
            scaled_etf = scaled_etf * etf_scale_factor
            #print(scaled_etf)
            masked_etf = np.where(cmask == 0, scaled_etf, np.nan)
            masked_etf = np.where(masked_etf == -0.999900, np.nan, masked_etf)
            #print(masked_etf)

            # save the masked raster
            outetf = os.path.join(etfiledir, path_row, 'ETf', 'etf_' + date + '.tif')
            with rasterio.open(outetf, 'w', **outmeta) as wrast:
                wrast.write(masked_etf)

        print(f'Created ETf raster for {date}')

    # ETa raster
        # open file for geoproperties to use for outputting raster
        with rasterio.open(etaband) as src:
            outmeta = src.meta
            outmeta.update(dtype=rasterio.float32, compress='lzw')
            #apply scalar and cloud mask
            scaled_eta = src.read()
            scaled_eta = scaled_eta * eta_scale_factor
            #print(scaled_eta)
            masked_eta = np.where(cmask == 0, scaled_eta, np.nan)
            masked_eta = np.where(masked_eta == -9.999, np.nan, masked_eta)
            #print(masked_eta)

        # # save the masked raster
        # outeta = os.path.join(etfiledir, path_row, 'ETa', 'eta_' + date + '.tif')
        # with rasterio.open(outeta, 'w', **outmeta) as wrast:
        #     wrast.write(masked_eta)