import glob
import numpy as np
import os
import shutil
import sys
from datetime import datetime, timedelta
import rasterio
from rasterio.merge import merge
from pathlib import Path
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
import rasterio.shutil as rio_shutil


def warp_based_on_sample(input_raster: str, temp_folder: str, nodata=None, sample_file=None,
                         resamplemethod='average', outdtype='float32', generic_temp=False):

    if generic_temp:
        outname = 'temp.tif'
    else:
        outname = os.path.basename(input_raster)

    with rasterio.open(sample_file) as src:
        out_meta = src.meta
        crs = out_meta['crs']
        transform = out_meta['transform']
        cols = out_meta['width']
        rows = out_meta['height']
        # return out_meta

    if resamplemethod == 'nearest':
        rs = Resampling.nearest
    elif resamplemethod == 'average':
        rs = Resampling.average
    elif resamplemethod == 'bilinear':
        rs = Resampling.bilinear
    else:
        print('only nearest-neighbor, average and bi-linear resampling are supported at this time')
        sys.exit(0)

    with rasterio.open(input_raster, 'r') as src:
        # create the virtual raster based on the standard rasterio
        # attributes from the sample tiff and shapefile feature.
        # update with suitable nodata values.
        if nodata is not None:
            print(f'nodata set to {nodata}')
            with WarpedVRT(src, resampling=rs,
                           crs=crs,
                           transform=transform,
                           height=rows,
                           width=cols,
                           nodata=nodata,
                           dtype=outdtype) as vrt:
                # save the file as an enumerated tiff. reopen outside this loop with the outputs list
                outwarp = os.path.join(temp_folder, outname)
                rio_shutil.copy(vrt, outwarp, driver='GTiff')
        else:
            print('Nodata is not set')
            with WarpedVRT(src, resampling=rs,
                           crs=crs,
                           transform=transform,
                           height=rows,
                           width=cols,
                           dtype=outdtype) as vrt:
                # save the file as an enumerated tiff. reopen outside this loop with the outputs list
                outwarp = os.path.join(temp_folder, outname)
                rio_shutil.copy(vrt, outwarp, driver='GTiff')

    return outwarp


# https://stackoverflow.com/questions/77913144/merge-rasters-together-with-overlapping-extents-without-losing-the-pixels-in-the

def mutual_extent(etf_dir, mosaic_dir):

    rasters = sorted(glob.glob(os.path.join(etf_dir, '*.tif')))
    print(rasters)

    sample_raster = rasters[0]
    rasters_to_warp = rasters[1:]
    # Set up a list to hold the aligned rasters
    aligned_rasters = [sample_raster]  # starts of with the sample raster because it matters too!
    for ras in rasters_to_warp:
        warp_ras = warp_based_on_sample(ras, temp_folder=mosaic_dir, nodata=-9999, sample_file=sample_raster,
                                        resamplemethod='nearest', outdtype='float32')
        aligned_rasters.append(warp_ras)

    return aligned_rasters

def scale_raster(etf_dir, scale_dir, mutual_raster):
    os.chdir(etf_dir)
    epath = Path(etf_dir)
    rasters = list(epath.glob('*.tif'))
    print(rasters)
    for ras in rasters:
        with rasterio.open(ras) as src, rasterio.open(mutual_raster) as mras:
            outmeta = mras.meta
            #outmeta.update(dtype=rasterio.int16, compress='lzw')
            # scale values
            scaled_etf = src.read()
            scaled_etf = scaled_etf * 10000

        # Write the scaled raster to a new file
        outetf = os.path.join(scale_dir, str(os.path.basename(ras)))
        print(outetf)
        with rasterio.open(outetf, 'w', **outmeta) as wrast:
            wrast.write(scaled_etf)

    print(f'Scaled and interized ETf rasters, saved under {scale_dir}')


def gapfilling(rfile, int_file_raster_list, testing=False):
    targetraster = rfile
    print('targetraster: ', targetraster)
    name = os.path.basename(rfile)[-12:-4]
    print(name)
    year = name[:4]
    month = name[4:6]
    day = name[6:8]
    print(f'Year: {year}, Month: {month}, Day:{day}')
    scenedate = year + month + day
    DATE_FORMAT = '%Y%m%d'
    targetdate = datetime.strptime(scenedate, DATE_FORMAT)
    targetdateInt = int(targetdate.strftime(DATE_FORMAT))
    start_date = (targetdate - timedelta(days=limit)).strftime(DATE_FORMAT)
    end_date = (targetdate + timedelta(days=limit)).strftime(DATE_FORMAT)
    print(f'Interpolation period between {str(start_date)} and {str(end_date)}')

    prevrasters = []
    afterrasters = []
    for r in int_file_raster_list:
        if int(os.path.basename(r)[-12:-4]) > int(start_date) and int(os.path.basename(r)[-12:-4]) < targetdateInt:
            prevrasters.append(r)
        if int(os.path.basename(r)[-12:-4]) < int(end_date) and int(os.path.basename(r)[-12:-4]) > targetdateInt:
            afterrasters.append(r)

    prevMax = len(prevrasters)
    print(f'There are {str(len(prevrasters))} potential Reference Arrays prior to the Target Date')
    afterMax = len(afterrasters)
    print(f'There are {str(afterMax)} potential Reference Arrays after the Target Date')
    prevrasters.sort(reverse=True)
    afterrasters.sort()

    if prevMax == 0 or afterMax == 0:
        print(f'Not enough Reference Rasters within the Max Limit')
        print(f'No interpolation for: {scenedate}')
        with rasterio.open(targetraster) as tras:
            outmeta = tras.meta.copy()
            outmeta.update(driver='GTiff', dtype='float32')
            tras1 = tras.read(1)
            #scaled = tras1 * 0.0001
            #scaled2 = np.where(scaled < 0, 0, scaled)

            outgfetf = os.path.join(outdir, 'gapfilled_ETf' + scenedate + '.tif')
            with rasterio.open(outgfetf, 'w', **outmeta) as wrast:
                wrast.write(tras1, 1)
        print("Finished with No Interpolation")

    else:
        combos = list(zip(prevrasters, afterrasters))
        print('length of combos', len(combos))
        print('the combos \n', combos)
        #count = 0
        lintCount = 1

        #linted_rasters = []
        #if count <= len(combos):
            #print(count, len(combos))
        for j, combo in enumerate(combos):
            #for combo in combos:
            raster1 = combo[0]
            print(raster1)
            raster2 = combo[1]
            print(raster2)

            prevDate = str(os.path.basename(combo[0])[-12:-4])
            refDate1 = datetime.strptime(prevDate, DATE_FORMAT)
            afterDate = str(os.path.basename(combo[1])[-12:-4])
            refDate2 = datetime.strptime(afterDate, DATE_FORMAT)
            timeDiff = int((refDate2 - refDate1).days)
            prevGap = int((targetdate - refDate1).days)
            afterGap = int((refDate2 - targetdate).days)
            print(f'Time Difference: {str(timeDiff)}')
            print(f'Time prior to Target Date: {str(prevGap)}')
            print(f'Time after Target Date:{str(afterGap)}')

            print(f'Starting Linear Interpolation Round: {str(lintCount)}')

            if j == 0:
                with rasterio.open(targetraster) as tras, rasterio.open(raster1) as r1, rasterio.open(raster2) as r2:
                    outmeta = tras.meta.copy()
                    outmeta.update(driver='GTiff', dtype='float32')

                    tras = tras.read(1)
                    r1 = r1.read(1)
                    ra1 = np.ma.masked_less(r1, 0)
                    r1shape = np.shape(ra1)
                    r2 = r2.read(1)
                    ra2 = np.ma.masked_less(r2, 0)
                    r2shape = np.shape(ra2)
                    #if testing:
                    #    with rasterio.open(os.path.join(outdir, f'tras{scenedate}.tif'), 'w', **outmeta) as wrast:
                    #        wrast.write(tras, 1)

                    if r1shape == np.shape(tras) and r2shape == np.shape(tras):
                        slope = (ra2 - ra1) / timeDiff
                        #if testing:
                        #    with rasterio.open(os.path.join(outdir, f'slope{scenedate}.tif'), 'w', **outmeta) as wrast:
                        #        wrast.write(slope, 1)
                        linInterp = ra1 + (slope * prevGap)
                        #if testing:
                        #    with rasterio.open(os.path.join(outdir, f'linInterp{scenedate}.tif'), 'w', **outmeta) as wrast:
                        #        wrast.write(linInterp, 1)
                        # np.copyto(tras, linInterp, 'safe', tras == -9999)
                        # tras[tras == -9999] = linInterp

                        tras[tras == -9999] = np.nan
                        ras = np.where(np.isnan(tras), linInterp, tras)
                        # ras = np.ma.filled(tras, -9999)
                        # ras_scaled = ras * 0.0001
                        ras_final = np.where(ras < 0, 0, ras)
                    else:
                        print(f'rasters are not in the same shape.')

                if len(combos) == 1:
                    # if there is only one combo you need to write out the final file here
                    # print('only one combo writing final')
                    out_gf_etf = os.path.join(outdir, 'gapfilled_ETf' + scenedate + '.tif')
                    with rasterio.open(out_gf_etf, 'w', **outmeta) as wrast:
                            wrast.write(ras_final, 1)
                else:
                    # print('more than one combo, writing first temp')
                    temp_gap = os.path.join(temploc, 'gap_temp'+scenedate+'.tif')
                    with rasterio.open(temp_gap, 'w', **outmeta) as wrast:
                        wrast.write(ras_final, 1)


                    #out_gf_etf = os.path.join(outdir, 'gapfilled_ETf'+scenedate+'.tif')
                    #with rasterio.open(out_gf_etf, 'w', **outmeta) as wrast:
                    #   wrast.write(ras_final, 1)


                #print(f'Linear Interpolation Round {str(lintCount)} - Complete!')
                #print('--------------------------------------------------------')
                #count += 1
                #lintCount += 1
            # the second+ one we need to overwrite the temp file with increasing gapfills
            elif (j > 0) and (j != (len(combos)-1)):
                print('doing middle')
                with rasterio.open(temp_gap) as tras, rasterio.open(raster1) as r1, rasterio.open(
                    raster2) as r2:
                    outmeta = tras.meta.copy()
                    outmeta.update(driver='GTiff', dtype='float32')

                    tras = tras.read(1)
                    r1 = r1.read(1)
                    ra1 = np.ma.masked_less(r1, 0)
                    r1shape = np.shape(ra1)
                    r2 = r2.read(1)
                    ra2 = np.ma.masked_less(r2, 0)
                    r2shape = np.shape(ra2)
                    # if testing:
                    #     with rasterio.open(os.path.join(outdir, f'tras{scenedate}.tif'), 'w', **outmeta) as wrast:
                    #         wrast.write(tras, 1)

                    if r1shape == np.shape(tras) and r2shape == np.shape(tras):
                        slope = (ra2 - ra1) / timeDiff
                        # if testing:
                        #     with rasterio.open(os.path.join(outdir, f'slope{scenedate}.tif'), 'w', **outmeta) as wrast:
                        #         wrast.write(slope, 1)
                        linInterp = ra1 + (slope * prevGap)
                        # if testing:
                        #     with rasterio.open(os.path.join(outdir, f'linInterp{scenedate}.tif'), 'w', **outmeta) as wrast:
                        #         wrast.write(linInterp, 1)
                        # np.copyto(tras, linInterp, 'safe', tras == -9999)
                        # tras[tras == -9999] = linInterp

                        tras[tras == -9999] = np.nan
                        ras = np.where(np.isnan(tras), linInterp, tras)
                        # ras = np.ma.filled(tras, -9999)
                        # ras_scaled = ras * 0.0001
                        ras_final = np.where(ras < 0, 0, ras)
                    else:
                        raise Exception(f'rasters are not in the same shape.')

                    # temp_file = os.path.join(outdir, 'gapfilled_ETf' + scenedate + '.tif')
                with rasterio.open(temp_gap, 'w', **outmeta) as wrast:
                    wrast.write(ras_final, 1)
                print('middle lint Complete')


            # last one we want to write the final one out.
            elif j == (len(combos)-1):
                print('doing end')
                with rasterio.open(temp_gap) as tras, rasterio.open(raster1) as r1, rasterio.open(
                    raster2) as r2:
                    outmeta = tras.meta.copy()
                    outmeta.update(driver='GTiff', dtype='float32')

                    tras = tras.read(1)
                    r1 = r1.read(1)
                    ra1 = np.ma.masked_less(r1, 0)
                    r1shape = np.shape(ra1)
                    r2 = r2.read(1)
                    ra2 = np.ma.masked_less(r2, 0)
                    r2shape = np.shape(ra2)
                    # if testing:
                    #     with rasterio.open(os.path.join(outdir, f'tras{scenedate}.tif'), 'w', **outmeta) as wrast:
                    #         wrast.write(tras, 1)

                    if r1shape == np.shape(tras) and r2shape == np.shape(tras):
                        slope = (ra2 - ra1) / timeDiff
                        # if testing:
                        #     with rasterio.open(os.path.join(outdir, f'slope{scenedate}.tif'), 'w', **outmeta) as wrast:
                        #         wrast.write(slope, 1)
                        linInterp = ra1 + (slope * prevGap)
                        # if testing:
                        #     with rasterio.open(os.path.join(outdir, f'linInterp{scenedate}.tif'), 'w', **outmeta) as wrast:
                        #         wrast.write(linInterp, 1)
                        # np.copyto(tras, linInterp, 'safe', tras == -9999)
                        # tras[tras == -9999] = linInterp

                        tras[tras == -9999] = np.nan
                        ras = np.where(np.isnan(tras), linInterp, tras)
                        # ras = np.ma.filled(tras, -9999)
                        # ras_scaled = ras * 0.0001
                        ras_final = np.where(ras < 0, 0, ras)
                    else:
                        print(f'rasters are not in the same shape.')

                    out_gf_etf = os.path.join(outdir, 'gapfilled_ETf' + scenedate + '.tif')
                    with rasterio.open(out_gf_etf, 'w', **outmeta) as wrast:
                        wrast.write(ras_final, 1)
                    print(f'wrote the final combo, Year: {year}, Month: {month}, Day:{day} is complete')
            # print(f'Linear Interpolation Round {str(lintCount)}  (final LINT) - Complete!')
            # print('ended combos wrote out ->', out_gf_etf)
            print('j', j, 'len combos', combos)
            print('--------------------------------------------------------')


if __name__ == '__main__':

    dirpath = input(f'Enter folder path to .../data/"pathrow" folder: ')
    os.chdir(dirpath)
    parentdir = os.path.normpath(os.getcwd() + os.sep + os.pardir)
    print(parentdir)
    pr = os.path.basename(dirpath)   #'p43r33'
    print(pr)
    # window of days to interpolate before and after the target image
    limit = 48
    cat = 'ETf'

    rasters = sorted(glob.glob(dirpath + os.sep + cat + os.sep + '*.tif'))
    print('me rasters: \n', rasters)

    startTime = datetime.now()
    # Set up output and temp directories
    outdir = os.path.join(dirpath, 'gapfilled' + cat)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    temp = os.path.join(dirpath, 'temp')
    if not os.path.exists(temp):
        os.makedirs(temp)

    outdir2 = os.path.join(temp, 'Int'+cat)
    if not os.path.exists(outdir2):
        os.makedirs(outdir2)
    temp2 = os.path.join(outdir2, 'temp')
    if not os.path.exists(temp2):
        os.makedirs(temp2)
    outdir3 = os.path.join(temp, 'MutualExtent')
    if not os.path.exists(outdir3):
        os.makedirs(outdir3)

    # Create the mutual extent raster for all scenes
    etf_dir = os.path.join(dirpath, cat)
    etf_rasters_aligned = mutual_extent(etf_dir=etf_dir, mosaic_dir=temp)
    print('aligned rasters: \n', etf_rasters_aligned)

    # clip ETf raster to mutual extent, scale and integerize raster to minimize file size
    startTime = datetime.now()
    mut_ras = os.path.join(temp, 'MutualExtent.tif')
    #scale_raster(etf_dir=etf_dir, scale_dir=outdir2, mutual_raster=mut_ras)
    InttimeEnd = str(datetime.now() - startTime)
    print(f'Total Time to Integerize: {str(InttimeEnd)}')

    # gapfilling the scaled ETf rasters using linear interpolation
    newstartTime = datetime.now()
    # list of integerized rasters
    # epath = Path(outdir2)
    # rasters = list(epath.glob('*.tif'))
    # print('rasters after whatever this is: ', rasters)

    for i in etf_rasters_aligned:
        print('i', i)
        gapfilling(rfile=i, int_file_raster_list=etf_rasters_aligned)

    print(f'ET Fraction has completed Linear Interpolation')

    LinttimeEnd = str(datetime.now() - newstartTime)
    print(f'Total Time to Interpolate {pr}: {str(LinttimeEnd)}')
