import glob
import os
from datetime import datetime
from datetime import timedelta
from pathlib import Path
import rasterio
import sys
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
import rasterio.shutil as rio_shutil

def warp_based_on_sample(input_raster: str, temp_folder: str, nodata=None, sample_file=None,
                         resamplemethod='average', outdtype='float32', generic_temp=False):

    if generic_temp:
        outname = 'temp.tif'
    else:
        outname = os.path.basename(input_raster)
    print('the sample file:', sample_file)

    with rasterio.open(sample_file) as src:
        out_meta = src.meta
        crs = out_meta['crs']
        transform = out_meta['transform']
        cols = out_meta['width']
        rows = out_meta['height']
        # return out_meta

    if resamplemethod == 'nearest':
        rs = Resampling.nearest
    elif resamplemethod == 'average':
        rs = Resampling.average
    elif resamplemethod == 'bilinear':
        rs = Resampling.bilinear
    else:
        print('only nearest-neighbor, average and bi-linear resampling are supported at this time')
        sys.exit(0)

    with rasterio.open(input_raster, 'r') as src:
        # create the virtual raster based on the standard rasterio
        # attributes from the sample tiff and shapefile feature.
        # update with suitable nodata values.
        if nodata is not None:
            print(f'nodata set to {nodata}')
            with WarpedVRT(src, resampling=rs,
                           crs=crs,
                           transform=transform,
                           height=rows,
                           width=cols,
                           nodata=nodata,
                           dtype=outdtype) as vrt:
                # save the file as an enumerated tiff. reopen outside this loop with the outputs list
                outwarp = os.path.join(temp_folder, outname)
                rio_shutil.copy(vrt, outwarp, driver='GTiff')
        else:
            print('Nodata is not set')
            with WarpedVRT(src, resampling=rs,
                           crs=crs,
                           transform=transform,
                           height=rows,
                           width=cols,
                           dtype=outdtype) as vrt:
                # save the file as an enumerated tiff. reopen outside this loop with the outputs list
                outwarp = os.path.join(temp_folder, outname)
                rio_shutil.copy(vrt, outwarp, driver='GTiff')

    return outwarp

prstartTime = datetime.now()
dirpath = input(f'Enter folder path to .../data/"pathrow" folder: ')
petdir = input(f'Enter folder path to reference ET data: ')

epath = os.path.join(dirpath, 'gapfilledETf')
rasters = glob.glob(os.path.join(epath, '*.tif'))
#rasters = sorted(glob.glob(dirpath + os.sep + 'gapfilledETF' + os.sep + '*.tif'))
print(f'Rasters: \n{rasters}')

etf_root = os.path.join(dirpath, 'dailyETf')
if not os.path.exists(etf_root):
    os.makedirs(etf_root)

eta_root = os.path.join(dirpath, 'dailyETa')
if not os.path.exists(eta_root):
    os.makedirs(eta_root)

temp_root = os.path.join(dirpath, 'temp')
# should already exist.
if not os.path.exists(eta_root):
    os.makedirs(eta_root)

for i in range(1, len(rasters[:-1])):
    current = rasters[i]
    nextone = rasters[i + 1]
    print(f'current raster: {os.path.basename(current)[-12:-4]}')
    print(f'next raster: {os.path.basename(nextone)[-12:-4]}')
    year = os.path.basename(current)[-12:-8]
    month = os.path.basename(current)[-8:-6]
    day = os.path.basename(current)[-6:-4]
    cur_date = datetime(year=int(year), month=int(month), day=int(day))
    next_date = datetime(year=int(os.path.basename(nextone)[-12:-8]),
                         month=int(os.path.basename(nextone)[-8:-6]),
                         day=int(os.path.basename(nextone)[-6:-4]))
    print(f'current date: {cur_date}')
    print(f'next date: {next_date}')

    # change file name string to match your reference et naming convention.
    # Here it is: pet_JJJ.tif, with JJJ = 3 digit day of the year
    pet_name = 'pet_' + datetime.strftime(cur_date, format='%j') + '.tif'
    pet = os.path.join(petdir, pet_name)
    print(f'pet raster: {pet}')

    # generic temp set to true because we dont' want to write out the file over and over in the temp folder.
    warped_pet = warp_based_on_sample(input_raster=pet, temp_folder=temp_root, sample_file=current, generic_temp=True)

    with rasterio.open(current) as curras_src, rasterio.open(nextone) as nextras_src, rasterio.open(warped_pet) as petras_src:

        # read the bands
        curras = curras_src.read(1)
        nextras = nextras_src.read(1)
        petras = petras_src.read(1)
        # copy metadata
        outmeta = curras_src.meta
        outmeta.update({'driver': 'GTiff', 'dtype': 'float32'})

        DATE_FORMAT = '%Y%m%d'
        currentdate = datetime.strptime(current[-12:-4], DATE_FORMAT)
        nextdate = datetime.strptime(nextone[-12:-4], DATE_FORMAT)
        currentdate = cur_date
        nextdate = next_date
        timeDiff = int((nextdate - currentdate).days)
        print('time Diff', timeDiff)

        targdate = str(currentdate.year) + str(currentdate.month).zfill(2) + str(currentdate.day).zfill(2)
        print(f'{targdate}')

        # Create daily ETf raster
        etfout = os.path.join(etf_root, f'dailyETf{targdate}.tif')
        with rasterio.open(etfout, 'w', **outmeta) as wras:
            wras.write(curras, 1)
        print(f'Created daily ETf {targdate}')

        # Create daily ETa raster
        etr = petras * 0.01
        eta = curras * etr
        etaout = os.path.join(dirpath, 'dailyETa', f'dailyETa{targdate}.tif')
        with rasterio.open(etaout, 'w', **outmeta) as wras:
            wras.write(eta, 1)
        print(f'Created daily ETa {targdate}')

        # full Interpolation step
        if timeDiff > 0:
            slope = (nextras - curras) / timeDiff
            for n in range(1, timeDiff):
                interpDate = currentdate + timedelta(days=n)
                print(f'interpolation date: {interpDate}')
                interpolateddate = str(interpDate.year) + str(interpDate.month).zfill(2) + str(interpDate.day).zfill(2)
                lint = curras + (slope * n)
                pet2 = os.path.join(petdir, 'pet_' + datetime.strftime(interpDate, format='%j') + '.tif')

                # generic temp set to true because we dont' want to write out the file over and over in the temp folder.
                warped_pet_2 = warp_based_on_sample(input_raster=pet2, temp_folder=temp_root, sample_file=current,
                                                    generic_temp=True)
                # Create daily interpolated ETf raster
                with rasterio.open(warped_pet_2) as petras2:
                    petras2_arr = petras2.read(1)
                    etr2 = petras2_arr * 0.01

                    outfile1 = os.path.join(etf_root, f'dailyETf{interpolateddate}.tif')
                    if os.path.isfile(outfile1) == False:
                        with rasterio.open(outfile1, 'w', **outmeta) as wras:
                            wras.write(lint, 1)
                    print(f'Created daily ETf {interpolateddate}')

                # Create daily interpolated ETf raster
                    interETa = lint * etr2
                    outfile2 = os.path.join(eta_root, f'dailyETa{interpolateddate}.tif')
                    with rasterio.open(outfile2, 'w', **outmeta) as wras:
                        wras.write(interETa, 1)
                    print(f'Created daily ETa {interpolateddate}')




