# Import system modules

from datetime import datetime
import glob
import os
import rasterio
import itertools
import calendar

def check_complete_year(yearly_eta_list, yr):
    first_day = parse_eta_date(yearly_eta_list[0])
    last_day = parse_eta_date(yearly_eta_list[-1])
    print(f'first day {first_day} and last day {last_day}')

    return first_day == datetime(year=yr, day=1, month=1) and last_day == datetime(year=yr, day=31, month=12)

def get_year_month_tuples(start_year, start_month, end_year, end_month):

    start_date = datetime(int(start_year), int(start_month), 1)
    end_date = datetime(int(end_year), int(end_month), 1)

    year_month_tuples = []
    current_date = start_date

    while current_date <= end_date:
        year_month_tuples.append((current_date.year, current_date.month))
        if current_date.month == 12:
            current_date = datetime(current_date.year + 1, 1, 1)
        else:
            current_date = datetime(current_date.year, current_date.month + 1, 1)

    return year_month_tuples

def parse_eta_date(eta_filepath):

    fname = os.path.basename(eta_filepath)
    dt_str = fname[8:16]
    print('dt_str:', dt_str)
    dt = datetime.strptime(dt_str, '%Y%m%d')
    return dt

def check_complete_month(monthly_eta_list):

    first_day = parse_eta_date(monthly_eta_list[0])
    last_day = parse_eta_date(monthly_eta_list[-1])
    print(f'first day: {first_day}, last day: {last_day}')

    try:
        # first day of month
        is_first = first_day.day == 1

        # last day of month
        ldom = calendar.monthrange(last_day.year, last_day.month)[1]
        is_last = last_day.day == ldom

        # returns True if it is the first and last day
        return is_first and is_last

    except ValueError:
        return False

def monthly_ETa(input_dir, year, output_directory):

    parent_dir = os.path.dirname(input_dir)
    outdir = os.path.join(parent_dir, 'monthlyETa')
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    eta_list = sorted(glob.glob(os.path.join(input_dir, '*.tif')))

    start_month = 1
    end_month = 12
    for i in range(start_month, (end_month + 1)):  # loop through month 1,2,..12,
        print('Month to be accumulated is: ' + str(i))
        Listras = []
        for et_path in eta_list:
            et_in = os.path.basename(et_path)
            if et_in[8:12] == str(year):
                mon = int(et_in[12:14]) # 2-digit month in daily ETa file name
                if mon == i:  # if month = i then append grid to list for summing up
                    Listras.append(et_path)
        print('rasters for month ' + str(i) + ': ')
        print(Listras)
        if Listras == []:
            print('No data for month ' + str(i) + ' available..continue to next month')
            continue
        elif not check_complete_month(Listras):
            print(f'data for month {i} is incomplete. Continue to next month')
            continue
        else:

            for j, ras in enumerate(Listras):

                if j == 0:
                    # open file for geoproperties to use for outputting raster
                    with rasterio.open(Listras[j], 'r') as src:
                        outmeta = src.meta
                        # instantiate the cumulative raster.
                        cum_arr = src.read(1)
                else:
                    with rasterio.open(Listras[j], 'r') as src:
                        daily_arr = src.read(1)
                        cum_arr += daily_arr

            # save the summed up ET raster
            prodname = f'eta_{year}{i:02d}.tif'
            output_Tif = os.path.join(output_directory, prodname)
            with rasterio.open(output_Tif, 'w', **outmeta) as wrast:
                wrast.write(cum_arr, 1)

def annual_ETa(input_dir, year):

    # the yearly rasters
    eta_list = sorted(glob.glob(os.path.join(input_dir, f'dailyETa{year}*.tif')))
    print(eta_list)

    parent_dir = os.path.dirname(input_dir)
    outdir = os.path.join(parent_dir, 'yearlyETa')

    if not check_complete_year(eta_list, int(year)):
        raise Exception(f'Incomplete set of daily rasters for year {year}, you must process '
                        f'more rasters prior to aggregating.')
            # open file for geoproperties to use for outputting raster
    else:
        for j, ras in enumerate(eta_list):

            if j == 0:
                # open file for geoproperties to use for outputting raster
                with rasterio.open(eta_list[j], 'r') as src:
                    outmeta = src.meta
                    # instantiate the cumulative raster.
                    cum_arr = src.read(1)
            else:
                with rasterio.open(eta_list[j], 'r') as src:
                    daily_arr = src.read(1)
                    cum_arr += daily_arr

            # save the summed up ET raster
            prodname = f'eta_{year}.tif'
            output_Tif = os.path.join(outdir, prodname)
            with rasterio.open(output_Tif, 'w', **outmeta) as wrast:
                wrast.write(cum_arr, 1)
        print('Created annual raster!')

def seasonal_ETa(input_dir, start_month, end_month, year):

    parent_dir = os.path.dirname(input_dir)
    outdir = os.path.join(parent_dir, 'seasonalETa')
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    number_of_month = (end_month+1) - start_month
    print(number_of_month)
    if start_month < end_month:
        # month_delta = end_month - start_month
        end_year = year
    elif start_month <= end_month:
        # we assume the season crosses calendar year.
        # first_tranche = 12 - start_month
        # month_delta = first_tranche + end_month
        end_year = year + 1

    yr_mo_lst = get_year_month_tuples(start_year=year, start_month=start_month, end_year=end_year, end_month=end_month)

    print('year month list: \n', yr_mo_lst)

    eta_list_of_lists = []
    for ym in yr_mo_lst:
        # unpack tuple
        y, m = ym
        etas_ym = sorted(glob.glob(os.path.join(parent_dir, 'dailyETa', f'dailyETa{y:04}{m:02}*.tif')))
        if not check_complete_month(etas_ym):
            raise Exception(f'Month {m} of year {y} does not have a complete set of daily ETa rasters, '
                            f'therefore the season cannot be aggregated. Try a different date range or '
                            f'adjust your interpolation.')
        eta_list_of_lists.append(etas_ym)

    eta_list = list(itertools.chain(*eta_list_of_lists))  # flattens the list of lists

    print('eta list of files for season: \n', eta_list)

    for j, ras in enumerate(eta_list):

        if j == 0:
            # open file for geoproperties to use for outputting raster
            with rasterio.open(eta_list[j], 'r') as src:
                outmeta = src.meta
                # instantiate the cumulative raster.
                cum_arr = src.read(1)
        else:
            with rasterio.open(eta_list[j], 'r') as src:
                daily_arr = src.read(1)
                cum_arr += daily_arr

    # save the summed up ET raster
    prodname = f'eta_{year}{start_month:02d}_{end_year}{end_month:02d}.tif'
    output_Tif = os.path.join(outdir, prodname)
    with rasterio.open(output_Tif, 'w', **outmeta) as wrast:
        wrast.write(cum_arr, 1)

if __name__ == '__main__':

        dird = r'W:\Users\Stefanie\IWMI\Awash_basin\L8_9_52%_169-166_051-054\2010_2024\052\data'
        time_period = 'monthly'
        # time_period = 'annual'
        # time_period = 'seasonal'
        for pat in ['p66r52', 'p67r52', 'p68r52']:
            print(pat)
            for year in range(2010, 2019):
                year = year  # input('Enter year (YYYY): ') # 2023
                path = os.path.join(dird, pat, 'dailyETa')
                #path = input('Enter folder path where daily ETa files are located: ')
                et_rasters = sorted(glob.glob(os.path.join(path, '*.tif')))
                #print(et_rasters)

                outdir = os.path.join(dird, pat, f'{time_period}ETa')
                if not os.path.exists(outdir):
                    os.makedirs(outdir)

                if time_period == 'monthly':
                    monthly_ETa(input_dir=path, year=year, output_directory=outdir)
                elif time_period == 'annual':
                    annual_ETa(input_dir=path, year=year)
                elif time_period == 'seasonal':
                    seasonal_ETa(input_dir=path, start_month=8, end_month=9, year=year)
                print('year done - next')